package daemon

import (
	"bufio"
	"context"
	"fmt"
	"net"
	"time"

	"github.com/open-io/oio-exporter/common"
	_log "github.com/open-io/oio-exporter/common/log"
)

type TcpDaemonFactory struct {
	serviceType string
}

func NewTcpDaemonFactory(serviceType string) DaemonFactoryI {
	return &TcpDaemonFactory{
		serviceType: serviceType,
	}
}

func (df *TcpDaemonFactory) Daemons(inventory *common.Inventory) map[string]interface{} {
	params := make(map[string]interface{})
	namespace := inventory.Namespace

	services := inventory.Services[df.serviceType]
	for _, service := range services {
		params[service.ID] = &tcpDaemonParams{
			Namespace:   namespace,
			ServiceID:   service.ID,
			ServiceIP:   service.IP,
			ServicePort: service.Port,
		}
	}

	return params
}

func (df *TcpDaemonFactory) New(paramsI interface{}) DaemonI {
	params := paramsI.(*tcpDaemonParams)

	return &TcpDaemon{
		Namespace:   params.Namespace,
		ServiceID:   params.ServiceID,
		ServiceIP:   params.ServiceIP,
		ServicePort: params.ServicePort,
	}
}

type tcpDaemonParams struct {
	Namespace   string
	ServiceID   string
	ServiceIP   string
	ServicePort int
}

type TcpDaemon struct {
	BaseDaemon

	Namespace   string
	ServiceID   string
	ServiceIP   string
	ServicePort int
	ServiceAddr string

	Timeout   time.Duration
	Interval_ time.Duration
	Keepalive bool
	Logger    *_log.Log

	conn net.Conn
}

func (d *TcpDaemon) Interval() time.Duration {
	return d.Interval_
}

func (d *TcpDaemon) Init(ctx context.Context) error {
	d.ServiceAddr = fmt.Sprintf("%s:%d", d.ServiceIP, d.ServicePort)

	return nil
}

// This should be overriden
func (d *TcpDaemon) Collect(ctx context.Context, m *common.MetricData) error {
	d.Logger.Fatal("Collect should be implemented when using the TcpDaemon")

	return nil
}

func (d *TcpDaemon) Command(ctx context.Context, cmd string, ends []string) ([]string, error) {
	target := fmt.Sprintf("%s:%d", d.ServiceIP, d.ServicePort)

	// retrieve connection from cache
	if conn := d.conn; conn != nil {
		data, err := d.write(ctx, conn, cmd, ends)
		if err == nil {
			return data, err
		}

		d.Logger.Warnf("connection from cache reported error, closing and reopening: %s", err)
		conn.Close()
		d.conn = nil
	}

	var dialer net.Dialer
	connCtx, cancel := context.WithTimeout(ctx, d.Timeout)
	defer cancel()

	conn, err := dialer.DialContext(connCtx, "tcp", target)
	if err != nil {
		return nil, err
	}
	if d.Keepalive {
		d.conn = conn
	} else {
		defer conn.Close()
	}

	return d.write(ctx, conn, cmd, ends)
}

func (d *TcpDaemon) write(ctx context.Context, conn net.Conn, cmd string, ends []string) ([]string, error) {
	if err := conn.SetDeadline(time.Now().Add(d.Timeout)); err != nil {
		err = fmt.Errorf("Error setting timeout writing command '%s' to '%s' (%w)", cmd, conn.RemoteAddr(), err)
		return nil, err
	}
	if _, err := conn.Write([]byte(cmd)); err != nil {
		err = fmt.Errorf("Error writing command '%s' to '%s' (%w)", cmd, conn.RemoteAddr(), err)
		return nil, err
	}

	data := []string{}

	scanner := bufio.NewScanner(conn)
	for scanner.Scan() {
		line := scanner.Text()
		for _, end := range ends {
			if line == end {
				return data, nil
			}
		}
		data = append(data, line)
	}
	return data, nil
}

func (d *TcpDaemon) Close() error {
	if d.conn == nil {
		return nil
	}

	return d.conn.Close()
}
