package main

import (
	"context"
	"errors"
	"flag"
	"fmt"
	"io"
	"net/http"
	"os"
	"os/signal"
	"reflect"
	"strings"
	"sync"
	"syscall"
	"time"

	"github.com/open-io/oio-exporter/beanstalkd"
	"github.com/open-io/oio-exporter/common"
	_log "github.com/open-io/oio-exporter/common/log"
	"github.com/open-io/oio-exporter/daemon"
	"github.com/open-io/oio-exporter/healthcheck"
	"github.com/open-io/oio-exporter/internal"
	clog "github.com/open-io/oio-exporter/log"
	"github.com/open-io/oio-exporter/memcached"
	"github.com/open-io/oio-exporter/oiofs"
	"github.com/open-io/oio-exporter/openio"
	"github.com/open-io/oio-exporter/redis"
	"github.com/open-io/oio-exporter/s3rt"
	"github.com/open-io/oio-exporter/static"
	"github.com/open-io/oio-exporter/version"
	"github.com/open-io/oio-exporter/zookeeper"
	"gopkg.in/yaml.v2"
)

var log = _log.Default

// Those are base64 encoding of their respective files
// They are filled through flags
// See the Makefile
var (
	DEFAULT_INVENTORY_PATH = "/etc/oio/sds/OPENIO/inventory.yml"
	HEALTHCHECKS           = ""
	LOGPATTERNS            = ""
	VERSIONS               = ""
)

var (
	host            string
	inventoryPath   string
	remoteExporters []string
	patternPath     string
	healthcheckPath string
	versionPath     string
	bind            string
	namespaceName   string
	prefix          string
	daemonsToRun    string
	daemonsNotToRun string
)

func parseFlags() {
	defaultHost, err := os.Hostname()
	if err != nil {
		log.Warnf("Could not retrieve hostname: %s", err)
	}

	remoteExportersStr := ""
	defaultNamespaceName, ok := os.LookupEnv("OIO_NS")
	if !ok {
		defaultNamespaceName = "OPENIO"
	}

	fs := flag.NewFlagSet("", flag.ExitOnError)
	fs.StringVar(&inventoryPath, "inventory", DEFAULT_INVENTORY_PATH, "Path to inventory file")
	fs.StringVar(&patternPath, "pattern-path", "", "Path to log patterns")
	fs.StringVar(&healthcheckPath, "healthcheck-path", "", "Path to healthchecks definition")
	fs.StringVar(&versionPath, "version-path", "", "Path to versions definition")
	fs.StringVar(&host, "host", defaultHost, "Report metrics with the specified hostname")
	fs.StringVar(&bind, "bind", "0.0.0.0:6920", "Listen address")
	fs.StringVar(&namespaceName, "namespace", defaultNamespaceName, "Namespace")
	fs.StringVar(&prefix, "prefix", "openio_", "Metrics prefix")
	fs.StringVar(&remoteExportersStr, "targets", "", "Comma separated list of targets to perform healthchecks on")
	fs.StringVar(&daemonsToRun, "daemons", "", "Comma separated list of daemons to run (empty means all of them)")
	fs.StringVar(&daemonsNotToRun, "nodaemons", "", "Comma separated list of daemons to explicitly disable (empty means nothing to disable)")

	_log.InitFlags(fs)
	daemon.InitFlags(fs)
	beanstalkd.InitFlags(fs)
	healthcheck.InitFlags(fs)
	internal.InitFlags(fs)
	memcached.InitFlags(fs)
	oiofs.InitFlags(fs)
	openio.InitFlags(fs)
	redis.InitFlags(fs)
	s3rt.InitFlags(fs)
	static.InitFlags(fs)
	version.InitFlags(fs)
	zookeeper.InitFlags(fs)

	if err := fs.Parse(os.Args[1:]); err != nil {
		log.Fatalf("FATAL Could not parse args: %s", err)
	}

	if err := _log.Setup(); err != nil {
		log.Fatalf("FATAL Failed to setup the logger: %s", err)
	}

	if len(remoteExportersStr) > 0 {
		remoteExporters = strings.Split(remoteExportersStr, ",")
	}

	allDaemonNames := "hc,oio,redis,version,s3rt,log,oiofs,internal,static,memcached,beanstalkd,zookeeper"
	if len(daemonsToRun) == 0 {
		daemonsToRun = allDaemonNames
	}
}

func main() {
	// Ideally we could use the `init` function, but it causes tests to fail from go 1.13
	// See https://golang.org/doc/go1.13#testing
	// And https://github.com/golang/go/issues/31859
	parseFlags()

	globalCtx, stopF := context.WithCancel(context.Background())
	metrics := common.MakeMetricData()

	allFactories := map[string]daemon.DaemonFactoryI{
		"beanstalkd": beanstalkd.Factory,
		"hc":         healthcheck.NewFactory(HEALTHCHECKS, healthcheckPath),
		"internal":   internal.Factory,
		"log":        clog.NewFactory(LOGPATTERNS, patternPath),
		"memcached":  memcached.Factory,
		"oiofs":      oiofs.Factory,
		"oio":        openio.Factory,
		"redis":      redis.Factory,
		"s3rt":       s3rt.Factory,
		"static":     static.Factory,
		"version":    version.NewFactory(VERSIONS, versionPath),
		"zookeeper":  zookeeper.Factory,
	}

	allRemoteFactories := map[string]struct{}{
		"hc": struct{}{},
	}

	factories := make(map[string]daemon.DaemonFactoryI)
	remoteFactories := make(map[string]daemon.DaemonFactoryI)
	for _, daemonName := range strings.Split(daemonsToRun, ",") {
		if common.Index(strings.Split(daemonsNotToRun, ","), daemonName) > -1 {
			log.Debugf("Ignoring daemon %s ...", daemonName)
			continue
		}

		factory, ok := allFactories[daemonName]
		if !ok {
			log.Fatalf("Unknown daemon '%s'", daemonName)
		}

		factories[daemonName] = factory

		if _, ok := allRemoteFactories[daemonName]; ok {
			remoteFactories[daemonName] = factory
		}
	}

	if len(factories) == 0 {
		log.Fatalf("no daemon selected, at least one is necessary to run.")
	}

	if len(remoteFactories) == 0 && len(remoteExporters) > 0 {
		remoteExporters = []string{}
		log.Info("targets have been disabled because none of the daemons specified use them")
	}

	daemons := make(map[string]*InventoryDaemons)
	daemons["_self_"] = NewInventoryDaemons(factories, globalCtx, metrics)
	for _, remoteExporter := range remoteExporters {
		daemons[remoteExporter] = NewInventoryDaemons(remoteFactories, globalCtx, metrics)
	}
	exporter := &OioExporter{
		inventoryPath: inventoryPath,
		ctx:           globalCtx,
		metrics:       metrics,
		daemons:       daemons,
		httpClient:    http.DefaultClient,
	}

	sigC := make(chan os.Signal, 1)
	signal.Notify(sigC, os.Interrupt, syscall.SIGTERM, syscall.SIGHUP)

	var wg sync.WaitGroup
	wg.Add(1 + 2*len(daemons))

	for _, inventoryDaemons := range daemons {
		go func(inventoryDaemons *InventoryDaemons) {
			defer wg.Done()
			inventoryDaemons.Run()
		}(inventoryDaemons)
	}

	// Reload periodically
	reloadEvery := 5 * time.Minute
	go func() {
		defer wg.Done()
		<-daemons["_self_"].readyC
		exporter.ReloadPeriodically(reloadEvery, "_self_", exporter.ReloadLocal)
	}()
	for _, remoteExporter := range remoteExporters {
		go func(remoteExporter string) {
			defer wg.Done()
			<-daemons[remoteExporter].readyC
			exporter.ReloadPeriodically(reloadEvery, remoteExporter, func() error {
				return exporter.ReloadRemote(remoteExporter)
			})
		}(remoteExporter)
	}

	go func() {
		defer wg.Done()
		<-daemons["_self_"].readyC // Otherwise reloading could crash
		Listen(globalCtx, host, bind, prefix, exporter)
	}()

	// TODO: handle SIGHUP to reload
	<-sigC

	log.Notice("Closing gracefully")
	// Cancel the global context
	stopF()

	// 20 seconds to stop gracefully

	waitC := make(chan struct{})
	go func() {
		defer close(waitC)

		wg.Wait()
	}()

	select {
	case <-waitC:
		log.Notice("Exited")
	case <-time.After(20 * time.Second):
		log.Fatal("Timeout")
	case <-sigC:
		log.Fatal("Forced Exit")
	}
}

func parseInventory(r io.Reader, namespaceName string) (*common.Inventory, error) {
	var mnInventory *common.MultiNamespacesInventory
	if err := yaml.NewDecoder(r).Decode(&mnInventory); err != nil {
		return nil, err
	}

	if mnInventory == nil {
		return nil, errors.New("Empty inventory")
	}

	namespace, ok := mnInventory.Namespaces[namespaceName]
	if !ok {
		return nil, fmt.Errorf("Namespace %s not found", namespaceName)
	}

	snInventory := &common.SingleNamespaceInventory{
		Namespace: namespaceName,
		Services:  namespace.Services,
		Localname: namespace.Localname,
		Rack:      namespace.Rack,
		Release:   mnInventory.Release,
		Hostname:  mnInventory.Hostname,
	}

	return (*common.Inventory)(snInventory), nil
}

type OioExporter struct {
	inventoryPath string

	ctx     context.Context
	metrics *common.MetricData

	daemons map[string]*InventoryDaemons

	httpClient *http.Client
}

type InventoryDaemons struct {
	ctx     context.Context
	metrics *common.MetricData

	factories map[string]daemon.DaemonFactoryI

	// Ready to handle the first reload
	readyC chan struct{}
	// Prevent concurrent reloading
	reloadC chan *common.Inventory
	// The params of daemons currently running
	daemonsState DaemonsState
	// Cancel functions to stop the daemons
	stopFunctions map[string]map[string]context.CancelFunc
	// Channels for the daemons to ack when they stopped
	ackCs map[string]map[string]chan struct{}

	// Keep track of all the goroutines launched
	wg sync.WaitGroup
}

func NewInventoryDaemons(factories map[string]daemon.DaemonFactoryI, ctx context.Context, metrics *common.MetricData) *InventoryDaemons {
	stopFunctions := make(map[string]map[string]context.CancelFunc)
	ackCs := make(map[string]map[string]chan struct{})
	for factoryID := range factories {
		stopFunctions[factoryID] = make(map[string]context.CancelFunc)
		ackCs[factoryID] = make(map[string]chan struct{})
	}
	return &InventoryDaemons{
		ctx:           ctx,
		metrics:       metrics,
		factories:     factories,
		readyC:        make(chan struct{}),
		reloadC:       make(chan *common.Inventory),
		daemonsState:  make(DaemonsState),
		stopFunctions: stopFunctions,
		ackCs:         ackCs,
	}
}

type DaemonsState map[string]map[string]interface{}

var ERR_ALREADY_RELOADING = errors.New("Concurrent reloads are not allowed")

// Reload the local inventory
func (exp *OioExporter) ReloadLocal() error {
	inventoryFile, err := os.Open(inventoryPath)
	if err != nil {
		return fmt.Errorf("Could not open inventory file: %w", err)
	}

	inventory, err := parseInventory(inventoryFile, namespaceName)
	inventoryFile.Close()
	if err != nil {
		return fmt.Errorf("Could not parse inventory: %w", err)
	}

	return exp.daemons["_self_"].reload(inventory)
}

// Reload the inventory from a remote exporter
func (exp *OioExporter) ReloadRemote(remoteExporter string) error {
	client := exp.httpClient

	url := fmt.Sprintf("http://%s/config", remoteExporter)
	req, err := http.NewRequestWithContext(exp.ctx, "GET", url, nil)
	if err != nil {
		return err
	}

	res, err := client.Do(req)
	if err != nil {
		return err
	}

	body := res.Body
	inventory, err := parseInventory(body, namespaceName)
	body.Close()
	if err != nil {
		return fmt.Errorf("Could not parse inventory: %w", err)
	}

	inventory.IsRemote = true

	return exp.daemons[remoteExporter].reload(inventory)
}

func (exp *OioExporter) ReloadPeriodically(reloadEvery time.Duration, id string, reloadFunc func() error) {
	ticker := time.NewTicker(reloadEvery)
	defer ticker.Stop()

	reloadC := ticker.C
	doneC := exp.ctx.Done()

	for {
		log.Noticef("Reloading '%s'", id)
		if err := reloadFunc(); err != nil {
			log.Warnf("Periodical reload failed for '%s': %s", id, err)
		}

		select {
		case <-doneC:
			return
		case <-reloadC:
		}
	}
}

func (id *InventoryDaemons) Run() {
	doneC := id.ctx.Done()

	close(id.readyC)

	var inventory *common.Inventory
	for {
		select {
		case <-doneC:
			goto exit
		default:
		}

		select {
		case <-doneC:
			goto exit
		case inventory = <-id.reloadC:
		}

		targetState := id.makeDaemonsState(inventory)
		id.matchDaemonsState(targetState)
		id.daemonsState = targetState
	}

exit:
	close(id.reloadC)
	id.wg.Wait()
}

func (id *InventoryDaemons) reload(inventory *common.Inventory) error {
	select {
	case id.reloadC <- inventory:
	default:
		return ERR_ALREADY_RELOADING
	}

	return nil
}

func (id *InventoryDaemons) makeDaemonsState(inventory *common.Inventory) DaemonsState {
	factories := id.factories
	targetState := make(DaemonsState)

	for factoryID, factory := range factories {
		if factory, ok := factory.(daemon.ReloaderDaemonFactoryI); ok {
			if err := factory.Reload(); err != nil {
				log.Warnf("Factory '%s' is broken: %s", factoryID, err)

				continue
			}
		}

		targetState[factoryID] = factory.Daemons(inventory)
	}

	return targetState
}

func (id *InventoryDaemons) matchDaemonsState(targetState DaemonsState) {
	exporterCtx := id.ctx
	factories := id.factories
	currentState := id.daemonsState

	// Daemons to start/reload
	for factoryID, factoryTargetState := range targetState {
		factoryCurrentState := currentState[factoryID]

		for daemonHash, targetDaemonParams := range factoryTargetState {
			daemonID := fmt.Sprintf("%s:%s", factoryID, daemonHash)
			currentDaemonParams, ok := factoryCurrentState[daemonHash]
			if ok {
				// No changes
				if reflect.DeepEqual(currentDaemonParams, targetDaemonParams) {
					continue
				}

				// Reload the daemon
				log.Infof("Daemon '%s' has to be reloaded", daemonID)

				// Stop it first
				stopF := id.stopFunctions[factoryID][daemonHash]
				ackC := id.ackCs[factoryID][daemonHash]
				id.stopDaemon(daemonID, stopF, ackC)
			}

			// Start daemon
			daemonCtx, stopF := context.WithCancel(exporterCtx)
			ackC := make(chan struct{})
			go id.startDaemon(daemonID, factories[factoryID], targetDaemonParams, daemonCtx, id.metrics, ackC)
			id.stopFunctions[factoryID][daemonHash] = stopF
			id.ackCs[factoryID][daemonHash] = ackC
		}
	}

	// Daemons to stop
	for factoryID, factoryCurrentState := range currentState {
		factoryTargetState := targetState[factoryID]

		for daemonHash := range factoryCurrentState {
			_, ok := factoryTargetState[daemonHash]

			if ok {
				continue
			}

			daemonID := fmt.Sprintf("%s:%s", factoryID, daemonHash)

			// Stop daemon
			stopF := id.stopFunctions[factoryID][daemonHash]
			ackC := id.ackCs[factoryID][daemonHash]
			go id.stopDaemon(daemonID, stopF, ackC)
		}
	}
}

func (id *InventoryDaemons) startDaemon(daemonID string, factory daemon.DaemonFactoryI, daemonParams interface{}, daemonCtx context.Context, metrics *common.MetricData, ackC chan<- struct{}) {
	d := factory.New(daemonParams)

	delay := d.Delay()

	log.Infof("Start daemon '%s' after waiting %s", daemonID, delay)
	time.Sleep(delay)

	id.wg.Add(1)
	defer id.wg.Done()

	if err := daemon.Run(daemonCtx, daemonID, d, metrics); err != nil {
		log.Errorf("Daemon '%s' failure: %s", daemonID, err)
	}

	close(ackC)
	log.Infof("Daemon '%s' has been stopped", daemonID)
}

func (id *InventoryDaemons) stopDaemon(daemonID string, stopF context.CancelFunc, ackC <-chan struct{}) {
	log.Infof("Stopping daemon '%s'", daemonID)
	stopF()

	<-ackC
}

func (id *InventoryDaemons) Close() {
	id.wg.Wait()
}
